package testcase;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;

import javax.resource.ResourceException;
import javax.resource.cci.ConnectionFactory;

import org.junit.AfterClass;
import org.junit.BeforeClass;
import org.junit.Test;
import org.teiid.adminapi.Model.Type;
import org.teiid.adminapi.impl.ModelMetaData;
import org.teiid.client.plan.PlanNode;
import org.teiid.jdbc.RequestOptions;
import org.teiid.jdbc.StatementCallback;
import org.teiid.jdbc.TeiidDriver;
import org.teiid.jdbc.TeiidStatement;
import org.teiid.resource.adapter.file.FileConnectionImpl;
import org.teiid.resource.adapter.ws.WSManagedConnectionFactory;
import org.teiid.runtime.EmbeddedConfiguration;
import org.teiid.runtime.EmbeddedServer;
import org.teiid.translator.FileConnection;
import org.teiid.translator.TranslatorException;
import org.teiid.translator.file.FileExecutionFactory;
import org.teiid.translator.ws.WSExecutionFactory;

public class BDMTestCase
{

    private static EmbeddedServer server;

    @SuppressWarnings("serial")
    private static final class FileConnectionFactory
        extends AbstractConnectionFactory<FileConnection>
    {
        @Override
        public FileConnection getConnection() throws ResourceException
        {
            return new FileConnectionImpl("src/test/resources", null, false);
        }
    }

    @BeforeClass
    public static void setUp() throws Exception
    {
        final EmbeddedConfiguration ec = new EmbeddedConfiguration();
        // ec.setUserRequestSourceConcurrency(1);
        // ec.setUseDisk(true);
        ec.setUserRequestSourceConcurrency(128);
        ec.setMaxActivePlans(128);
        ec.setMaxThreads(128);
        server = new EmbeddedServer();
        server.start(ec);

        final ConnectionFactory cf = new FileConnectionFactory();
        server.addConnectionFactoryProvider("source-file",
                                            new EmbeddedServer.ConnectionFactoryProvider<ConnectionFactory>()
                                            {
                                                @Override
                                                public ConnectionFactory getConnectionFactory() throws TranslatorException
                                                {
                                                    return cf;
                                                }
                                            });
        final WSManagedConnectionFactory mcf = new WSManagedConnectionFactory();
        server.addConnectionFactoryProvider("ws",
                                            new EmbeddedServer.ConnectionFactoryProvider<ConnectionFactory>()
                                            {
                                                @Override
                                                public ConnectionFactory getConnectionFactory() throws TranslatorException
                                                {
                                                    try
                                                    {
                                                        return mcf.createConnectionFactory();
                                                    } catch (final Exception e)
                                                    {
                                                        e.printStackTrace();
                                                    }
                                                    return null;
                                                }
                                            });

        server.addTranslator(new FileExecutionFactory());
        server.addTranslator(new TimerangeExecutionFactory());
        server.addTranslator(new WSExecutionFactory());
        buildDeployVDB(server);

    }

    @AfterClass
    public static void tearDown() throws Exception
    {

        server.stop();
    }

    @SuppressWarnings("unchecked")
    @Test
    public void testRefreshExressionJoin() throws Exception
    {
        final String sql = "SELECT WRAP.* FROM (SELECT * FROM chorus_time.timetable WHERE period = 10000) AS chorus_refresh_expression CROSS JOIN (SELECT CHORUS_B.* FROM (call bdm_rest.invokeHttp(action => 'GET', endpoint => (select RestURL FROM bdm_config_view.get_config where TenantId = 1))) AS CHORUS_B) AS WRAP";

        final TeiidDriver td = server.getDriver();
        final Connection c = td.connect("jdbc:teiid:Chorus", null);
        final Statement stmt = c.createStatement();
        final TeiidStatement tstatement = stmt.unwrap(TeiidStatement.class);
        generateQueryPlan(sql, stmt);
        tstatement.submitExecute(sql, new StatementCallback()
        {
            @Override
            public void onRow(final Statement s, final ResultSet rs) throws Exception
            {
                System.out.println("Row returned" + rs.getObject(1));
            }

            @Override
            public void onException(final Statement s, final Exception e) throws Exception
            {
                e.printStackTrace();
            }

            @Override
            public void onComplete(final Statement s) throws Exception
            {
                System.out.println("Complete");
            }
        },
                                 new RequestOptions().continuous(true));
        Thread.sleep(500000000);
        tstatement.cancel();

    }

    private static void buildDeployVDB(final EmbeddedServer teiidServer) throws Exception
    {
        final ModelMetaData mmd = new ModelMetaData();
        mmd.setName("ws");
        mmd.addSourceMapping("bdm_rest", "ws", "ws");

        final ModelMetaData fileModel = new ModelMetaData();
        fileModel.setName("bdm_file");
        fileModel.setSchemaSourceType("native");
        fileModel.addSourceMapping("text-connector", "file", "source-file");

        final ModelMetaData baseUrlViewModel = new ModelMetaData();
        baseUrlViewModel.setName("bdm_baseurl_view");
        baseUrlViewModel.setModelType(Type.VIRTUAL);
        baseUrlViewModel.setSchemaSourceType("ddl");
        baseUrlViewModel.setSchemaText(createBaseUrlView());

        final ModelMetaData timeModel = new ModelMetaData();
        timeModel.setName("chorus_time");
        timeModel.setSchemaSourceType("native");
        timeModel.addSourceMapping("chorus_time", "timerange-async", "");

        final ModelMetaData wsModel = new ModelMetaData();
        wsModel.setName("bdm_rest");
        wsModel.setSchemaSourceType("native");
        wsModel.addSourceMapping("bdm_rest", "ws", "ws");

        final ModelMetaData getConfigViewModel = new ModelMetaData();
        getConfigViewModel.setName("bdm_config_view");
        getConfigViewModel.setModelType(Type.VIRTUAL);
        getConfigViewModel.setSchemaSourceType("ddl");
        getConfigViewModel.setSchemaText(createGetConfigView());

        teiidServer.deployVDB("Chorus",

        wsModel, timeModel, getConfigViewModel, fileModel);

    }

    private static String createBaseUrlView() throws Exception
    {
        final StringBuilder result = new StringBuilder();
        result.append(" create view \"base_url\" as ");
        result.append(" select 'http://' || Host As RestURL, TenantId, Name from bdm_config_view.get_config; ");

        return result.toString();
    }

    private static String createGetConfigView() throws Exception
    {
        final StringBuilder result = new StringBuilder();
        result.append("create view \"get_config\" as SELECT 'http://' || Host As RestURL, TenantId, Name, Host, Port from "
                      + "(call bdm_file.getTextFiles('data/bdm-rest-config.txt')) f,"
                      + "TEXTTABLE(f.file COLUMNS TenantId    string, Name        string, Host        string, Port        string HEADER 1) T");

        return result.toString();
    }

    private void generateQueryPlan(final String sql, final Statement stmt) throws SQLException
    {
        stmt.execute("SET SHOWPLAN ON");
        stmt.execute("SET NOEXEC ON");
        stmt.execute(sql);
        final PlanNode p = ((TeiidStatement) stmt).getPlanDescription();
        System.out.println("Plan: " + p);
    }

}
