package testcase;

import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.Future;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicLong;
import java.util.concurrent.atomic.AtomicReference;

import org.apache.log4j.Logger;
import org.teiid.language.ColumnReference;
import org.teiid.language.Command;
import org.teiid.language.Comparison;
import org.teiid.language.DerivedColumn;
import org.teiid.language.Literal;
import org.teiid.language.QueryExpression;
import org.teiid.language.visitor.AbstractLanguageVisitor;
import org.teiid.translator.DataNotAvailableException;
import org.teiid.translator.ExecutionContext;
import org.teiid.translator.ResultSetExecution;
import org.teiid.translator.ReusableExecution;
import org.teiid.translator.TranslatorException;

import com.google.common.collect.Lists;
import com.google.common.util.concurrent.Futures;

public class TimerExecution
    implements
        ResultSetExecution,
        ReusableExecution<Object>
{

    public static final Logger                        LOGGER           = Logger.getLogger(TimerExecution.class);

    protected final AtomicReference<ExecutionContext> executionContext = new AtomicReference<>();

    private final AtomicLong                          nextStartTime    = new AtomicLong();

    final AtomicLong                                  nextEndTime      = new AtomicLong();

    enum State {
        RESULT, WAIT
    }

    final AtomicReference<State>              state = new AtomicReference<>(State.RESULT);

    protected final Long                      period;

    protected final Future<?>                 future;

    private final ArrayList<Callable<Object>> columnFns;

    private int                               limit;

    public TimerExecution(final QueryExpression command,
                          final ExecutionContext executionContext,
                          final ScheduledExecutorService executor)
    {
        final Long[] longHolder = new Long[1];
        command.getProjectedQuery()
               .getWhere()
               .acceptVisitor(new AbstractLanguageVisitor()
               {
                   @Override
                   public void visit(final Comparison comparison)
                   {
                       final ColumnReference leftExpression = (ColumnReference) comparison.getLeftExpression();
                       if (!leftExpression.getMetadataObject()
                                          .getCanonicalName()
                                          .equals(TimerangeExecutionFactory.PERIOD))
                       {
                           throw new IllegalArgumentException("where condition on "
                                                              + leftExpression
                                                              + " not allowed");
                       }
                       final Literal rightExpression = (Literal) comparison.getRightExpression();
                       longHolder[0] = (Long) rightExpression.getValue();
                   }
               });
        this.columnFns = Lists.newArrayList();
        for (final DerivedColumn colRef : command.getProjectedQuery()
                                                 .getDerivedColumns())
        {
            final String canonicalName = ((ColumnReference) colRef.getExpression()).getMetadataObject()
                                                                                   .getCanonicalName();
            switch (canonicalName)
            {
            case TimerangeExecutionFactory.PERIOD:
                this.columnFns.add(new Callable<Object>()
                {

                    @Override
                    public Object call() throws Exception
                    {
                        return TimerExecution.this.period;
                    }
                });
                break;
            case TimerangeExecutionFactory.START:
                this.columnFns.add(new Callable<Object>()
                {

                    @Override
                    public Object call() throws Exception
                    {
                        return new Timestamp(TimerExecution.this.nextStartTime.get());
                    }
                });
                break;
            case TimerangeExecutionFactory.STOP:
                this.columnFns.add(new Callable<Object>()
                {

                    @Override
                    public Object call() throws Exception
                    {
                        return new Timestamp(TimerExecution.this.nextEndTime.get());
                    }
                });
                break;
            case TimerangeExecutionFactory.ACTUAL_PERIOD:
                this.columnFns.add(new Callable<Object>()
                {

                    @Override
                    public Object call() throws Exception
                    {
                        return TimerExecution.this.nextEndTime.get()
                               - TimerExecution.this.nextStartTime.get();
                    }
                });
                break;
            case TimerangeExecutionFactory.EXECUTION_COUNT:
                this.columnFns.add(new Callable<Object>()
                {

                    @Override
                    public Object call() throws Exception
                    {
                        return TimerExecution.this.executionContext.get()
                                                                   .getCommandContext()
                                                                   .getReuseCount();
                    }
                });
                break;

            }
        }
        this.period = longHolder[0];
        this.executionContext.set(executionContext);
        this.nextEndTime.set(System.currentTimeMillis());
        this.nextStartTime.set(this.nextEndTime.get() - this.period);
        if (executionContext.getCommandContext().isContinuous())
        {
            this.future = executor.scheduleAtFixedRate(new Runnable()
            {

                @Override
                public void run()
                {
                    synchronized (TimerExecution.this)
                    {
                        TimerExecution.this.nextEndTime.set(System.currentTimeMillis());
                        TimerExecution.this.state.set(State.RESULT);
                        TimerExecution.this.executionContext.get()
                                                            .dataAvailable();
                    }
                }
            },
                                                       this.period,
                                                       this.period,
                                                       TimeUnit.MILLISECONDS);
        } else
        {
            this.future = Futures.immediateFuture(null);
        }
        this.limit = command.getLimit() == null
                                               ? Integer.MAX_VALUE
                                               : command.getLimit()
                                                        .getRowLimit();

    }

    @Override
    synchronized public void execute() throws TranslatorException
    {
        switch (this.state.get())
        {
        case RESULT:
            break;
        case WAIT:
            throw DataNotAvailableException.NO_POLLING;
        }
    }

    @Override
    synchronized public List<?> next() throws TranslatorException,
                                      DataNotAvailableException
    {
        final State state_snapshot = this.state.get();
        try
        {
            switch (state_snapshot)
            {
            case RESULT:
                if (this.limit >= 1)
                {
                    final ArrayList<Object> result = Lists.newArrayList();
                    for (final Callable<Object> fn : this.columnFns)
                    {
                        result.add(fn.call());
                    }
                    this.nextStartTime.set(this.nextEndTime.get());

                    return result;
                } else
                {
                    return null;
                }
            case WAIT:
                return null;
            default:
                throw new IllegalStateException(this.state.toString());
            }
        } catch (final Exception e)
        {
            throw new TranslatorException(e);
        } finally
        {
            switch (state_snapshot)
            {
            case RESULT:
                this.state.set(State.WAIT);
                break;
            case WAIT:
                break;
            }
        }
    }

    @Override
    public void close()
    {

    }

    @Override
    public void cancel() throws TranslatorException
    {
        this.future.cancel(true);
    }

    @Override
    public void reset(final Command c,
                      final ExecutionContext executionContext,
                      final Object connection)
    {
        // System.out.println("reset");
        this.executionContext.set(executionContext);
    }

    @Override
    public void dispose()
    {
        this.future.cancel(true);
    }

}
